/* 
    GTVscreen Widget for v4l frame grabber driver

    Copyright (C) 1999 Marcus Metzler (mocm@metzlerbros.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * gtvscreen.c 
 */

#include "gtvscreen.h"
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <sys/wait.h>
#include "ringbuffy.h"

#ifdef PTHREADS
#include <pthread.h>
#endif

#ifdef HAVE_LIBXXF86VM
#include <X11/extensions/xf86vmode.h>
#endif

char *combost[] =
    { "1/2", "2/3", "3/4", "4/5", "5/6", "6/7", "7/8", "8/9", "auto", "V",
	"H", "Analog", "DVBS", "DVBT", "DVBC", "notset"
};

enum {
	CHINPUT,
	CHCAPTURE,
	CHOVERLAY,
	CHATTRIB,
	CHFREQ,
	CHAUDIO,
	CHPLAY,
	CHSTOP,
	CHPAUSE,
	CHRECORD,
	CHFFWD,
	CHFBWD,
	CHFSKP,
	CHBSKP,
	LAST_SIGNAL
};

static int get_clip(Gtvscreen * gtv);
static void gtvscreen_realize();
static void gtvscreen_init();
guint gtvscreen_get_type();
static void gtk_gtvscreen_size_allocate();
static void gtvscreen_destroy();
static void set_capture();
static GdkFilterReturn pConfigFilter();
static gint pConfig();
static void set_geo();
static void setmaxsize();
static void gtvscreen_send_configure();
static gint paint();
static void grab_image();
static unsigned int which_format();
static gint button_press();
static GtkWidget *CreateNote();
static void gtvscreen_all_new(Gtvscreen *gtv);
#ifdef HAVE_LIBXXF86VM
static int toggle_fullscreen(int n, Gtvscreen * gtv);
#endif
static XErrorHandler xerror(); 

static GtkWidgetClass *parent_class = NULL;
static Boolean Move = FALSE;
static Window lastrwin;

static const gint MAXX = 1024;
static const gint MAXY = 768;

#define TIMEOUT 400

static guint gtvscreen_signals[LAST_SIGNAL] = { 0 };

#ifdef PTHREADS
pthread_mutex_t mutexctl = PTHREAD_MUTEX_INITIALIZER;
#endif
static int myioctl(int fd, int call, void *pointer)
{
	int status;

#ifdef PTHREADS
	pthread_mutex_lock(&mutexctl);
#endif

#ifndef NOCTL
	status = ioctl(fd, call, pointer);
#endif
#ifdef PTHREADS
	pthread_mutex_unlock(&mutexctl);
#endif

	return status;
}

XErrorHandler xerror(Display *dpy, XErrorEvent *event)
{
  switch (event->error_code){
  case BadWindow:
#ifdef DEBUG
    printf("Bad Window ingnored\n");
#endif
    return NULL;
    break;

  case BadMatch:
#ifdef DEBUG
    printf("BadMatch\n");
#endif
    return NULL;
    break;
 
  case BadColor:
#ifdef DEBUG
    printf("BadColor\n");
#endif
    return NULL;
    break;
    
    default:
      return NULL;
  }
}                    

static void copyfreq(freqtype * out, freqtype * in)
{
	out->fre = in->fre;
	out->sr = in->sr;
	out->lnb = in->lnb;
	out->fec = in->fec;
	out->type = in->type;
	out->pol = in->pol;
	out->apid = in->apid;
	out->vpid = in->vpid;
}

static int freqchange(freqtype * a, freqtype * b)
{
	return ((a->fre != b->fre) ||
		(a->sr != b->sr) ||
		(a->lnb != b->lnb) ||
		(a->fec != b->fec) ||
		(a->type != b->type) ||
		(a->pol != b->pol) ||
		(a->apid != b->apid) || (a->vpid != b->vpid));
}

static void gtvscreen_class_init(GtvscreenClass * klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = (GtkObjectClass *) klass;
	widget_class = (GtkWidgetClass *) klass;

	parent_class = gtk_type_class(gtk_widget_get_type());


	gtvscreen_signals[CHINPUT] =
	    gtk_signal_new("chinput",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chinput),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHCAPTURE] =
	    gtk_signal_new("chcapture",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chcapture),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHOVERLAY] =
	    gtk_signal_new("choverlay",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, choverlay),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHATTRIB] =
	    gtk_signal_new("chattrib",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chattrib),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHFREQ] =
	    gtk_signal_new("chfreq",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chfreq),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHAUDIO] =
	    gtk_signal_new("chaudio",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chaudio),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHPLAY] =
	    gtk_signal_new("chplay",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chplay),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHSTOP] =
	    gtk_signal_new("chstop",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chstop),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHPAUSE] =
	    gtk_signal_new("chpause",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chpause),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHRECORD] =
	    gtk_signal_new("chrecord",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chrecord),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);


	gtvscreen_signals[CHFFWD] =
	    gtk_signal_new("chffwd",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chffwd),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

	gtvscreen_signals[CHFBWD] =
	    gtk_signal_new("chfbwd",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chfbwd),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);


	gtvscreen_signals[CHFSKP] =
	    gtk_signal_new("chfskp",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chfskp),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);


	gtvscreen_signals[CHBSKP] =
	    gtk_signal_new("chbskp",
			   GTK_RUN_FIRST,
			   object_class->type,
			   GTK_SIGNAL_OFFSET(GtvscreenClass, chbskp),
			   gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);


	gtk_object_class_add_signals(object_class, gtvscreen_signals,
				     LAST_SIGNAL);

	object_class->destroy = gtvscreen_destroy;
	widget_class->size_allocate = gtk_gtvscreen_size_allocate;
	widget_class->realize = gtvscreen_realize;

	klass->chinput   = NULL;	// I dont need to know yet.
	klass->chcapture = NULL;
	klass->choverlay = NULL;
	klass->chattrib  = NULL;
	klass->chfreq    = NULL;
	klass->chaudio   = NULL;
	klass->chplay    = NULL;
	klass->chstop    = NULL;
	klass->chpause   = NULL;
	klass->chrecord  = NULL;
	klass->chffwd    = NULL;
	klass->chfbwd    = NULL;
	klass->chfskp    = NULL;
	klass->chbskp    = NULL;
}



static void gtk_gtvscreen_size_allocate(GtkWidget * widget,
					GtkAllocation * allocation)
{
	g_return_if_fail(widget != NULL);
	g_return_if_fail(IS_GTVSCREEN(widget));
	g_return_if_fail(allocation != NULL);

	widget->allocation = *allocation;

	if (GTK_WIDGET_REALIZED(widget)) {
		gdk_window_move_resize(widget->window,
				       allocation->x, allocation->y,
				       allocation->width,
				       allocation->height);

		gtvscreen_send_configure(GTVSCREEN(widget));
	}
}

void gtk_gtvscreen_size(Gtvscreen * gtv, gint width, gint height)
{
	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	GTK_WIDGET(gtv)->requisition.width = width;
	GTK_WIDGET(gtv)->requisition.height = height;
}


static gint refresh_timer(gpointer data)
{
	Gtvscreen *gtv;

#ifndef NOCARD
	int scr, depth;
	XSetWindowAttributes xswa;
	unsigned long mask;
	Window rwin, rootW;
	Display *disp;
	int x, y, width, height;

	gtv = (Gtvscreen *) data;

	x = gtv->vwin.x;
	y = gtv->vwin.y;

	width = gtv->vwin.width;
	height = gtv->vwin.height;


#ifdef DEBUG
	printf("REFRESH\n");
#endif
	if (!gtv->OVL_ON)
		return 0;
	disp = GDK_DISPLAY();
	scr = DefaultScreen(disp);
	rootW = RootWindow(disp, scr);
	depth = gtv->vbuf.depth;

	xswa.override_redirect = TRUE;
	xswa.backing_store = NotUseful;
	xswa.save_under = FALSE;
	xswa.background_pixel = BlackPixel(disp, scr);
	xswa.backing_store = NotUseful;
	mask = (CWBackPixel | CWSaveUnder | CWBackingStore |
		CWOverrideRedirect);
	if (Move) {
		rwin = XCreateWindow(disp, rootW, 0, 0,
				     gtv->vbuf.width, gtv->vbuf.height, 0,
				     CopyFromParent, InputOutput,
				     CopyFromParent, mask, &xswa);
		Move = FALSE;
	} else {
		rwin = XCreateWindow(disp, rootW,
				     x - 20, y - 20,
				     width + 20, height + 20,
				     0, CopyFromParent, InputOutput,
				     CopyFromParent, mask, &xswa);
	}
	lastrwin = rwin;
	XMapWindow(disp, rwin);
	XUnmapWindow(disp, rwin);
	XDestroyWindow(disp, rwin);
#endif
	gtv->refresh = 1;
	return 0;
}

static void refresh(gpointer data)
{
	Gtvscreen *gtv;
	
	gtv = (Gtvscreen *) data;

	if (gtv->refresh || !Move){
		gtv->refresh = 0;
		gtv->timeout = 	gtk_timeout_add(
			TIMEOUT,refresh_timer,(gpointer)gtv);
	}
}

guint gtvscreen_get_type()
{
	static guint gtv_type = 0;

	if (!gtv_type) {
		GtkTypeInfo gtv_info = {
			"Gtvscreen",
			sizeof(Gtvscreen),
			sizeof(GtvscreenClass),
			(GtkClassInitFunc) gtvscreen_class_init,
			(GtkObjectInitFunc) gtvscreen_init,
			(GtkArgGetFunc) NULL,
			(GtkArgSetFunc) NULL,
		};

		gtv_type = gtk_type_unique(gtk_vbox_get_type(), &gtv_info);
	}

	return gtv_type;
}

GtkWidget *gtvscreen_new()
{
	return gtvscreen_new_with_dev((gchar *) NULL);
}

GtkWidget *gtvscreen_new_with_dev(gchar *devname)
{
	GtkWidget *gtvw;
	Gtvscreen *gtv;

	gtvw = GTK_WIDGET(gtk_type_new(gtvscreen_get_type()));
	gtv = GTVSCREEN(gtvw);

	if (devname) {
		printf("setting dev to %s\n", devname);
		strcpy(gtv->devname, devname);
	} else {
		printf("setting dev to %s\n", "/dev/video");
		strcpy(gtv->devname, "/dev/video");
	}

	if ((gtv->fbttv = open(gtv->devname, O_RDWR)) < 0) {
		fprintf(stderr, "Cannot open %s\n", gtv->devname);
		exit(1);
	}

	gtvscreen_all_new(gtv);
	return gtvw;
}


GtkWidget *gtvscreen_new_with_fd(int fd)
{
	GtkWidget *gtvw;
	Gtvscreen *gtv;

	gtvw = GTK_WIDGET(gtk_type_new(gtvscreen_get_type()));
	gtv = GTVSCREEN(gtvw);
	gtv->fbttv = fd;
	printf("setting file descriptor for device to %d\n", gtv->fbttv);
	gtvscreen_all_new(gtv);
	return gtvw;
}

static void gtvscreen_all_new(Gtvscreen *gtv){
	unsigned char *tm;
	int msize;

#ifndef NOCARD
	if (myioctl(gtv->fbttv, VIDIOCGWIN, &(gtv->vwin)))
		perror("get window");
	if (myioctl(gtv->fbttv, VIDIOCGFBUF, &(gtv->vbuf)))
		perror("get frame buffer");
	if (myioctl(gtv->fbttv, VIDIOCGTUNER, &(gtv->vtuner)))
		perror("get tuner");
	/* Alloc memory for snapshot */
	if (myioctl(gtv->fbttv, VIDIOCGMBUF, &(gtv->mbuf)))
		perror("get mbuf");
	myioctl(gtv->fbttv, VIDIOCGCAP, &(gtv->vcap));

	gtv->dvb = ((gtv->vcap.type & VID_TYPE_DVB) ? 1:0); 
	if (!gtv->vcap.type & VID_TYPE_OVERLAY)
		gtv->remote = 1;
	msize = gtv->mbuf.size;
	fprintf(stderr, "msize: 0x%08x\n", msize);
	tm = (unsigned char *) mmap(0, msize, PROT_READ | PROT_WRITE,
				    MAP_SHARED, gtv->fbttv, 0);
	if (tm == (unsigned char *) -1) {
		fprintf(stderr, "Cannot mmap memory for snapshot\n");
		exit(1);
	}
#ifdef DEBUG
	fprintf(stderr, "tm: 0x%08x\n", (int) tm);
#endif
	gtv->TargetMem = tm;

#endif

}



#ifndef NOCARD
static Boolean onscreen(Gtvscreen * gtv, XWindowAttributes wts)
{
	int x1, x2, y1, y2;
	int wx1, wx2, wy1, wy2;
	GtkWidget *widget;
	Boolean xlap, ylap;

	widget = GTK_WIDGET(gtv);

	if (!(wts.map_state & IsViewable))
		return FALSE;

	x1 = gtv->vwin.x;
	y1 = gtv->vwin.y;
	x2 = x1 + gtv->vwin.width;
	y2 = y1 + gtv->vwin.height;

	wx1 = wts.x;
	wy1 = wts.y;
	wx2 = wx1 + wts.width + 2 * wts.border_width - 1;
	wy2 = wy1 + wts.height + 2 * wts.border_width - 1;

	xlap = ((wx1 < x1 && wx2 > x1) || (wx1 > x1 && wx1 < x2));
	ylap = ((wy1 < y1 && wy2 > y1) || (wy1 > y1 && wy1 < y2));

	if (xlap && ylap)
		return TRUE;
	else
		return FALSE;

}

#endif

static void gtvscreen_send_configure(Gtvscreen * gtv)
{
	GtkWidget *widget;
	GdkEventConfigure event;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));


	widget = GTK_WIDGET(gtv);

	event.type = GDK_CONFIGURE;
	event.window = widget->window;
	event.send_event = TRUE;
	event.x = widget->allocation.x;
	event.y = widget->allocation.y;
	event.width = widget->allocation.width;
	event.height = widget->allocation.height;

	gtk_widget_event(widget, (GdkEvent *) & event);
	set_geo(gtv);
}

/* Realize */
static void gtvscreen_realize(GtkWidget * widget)
{
	Gtvscreen *gtv;
	GdkWindowAttr attributes;
	gint attributes_mask;
	uint rwidth;
	GdkWindow *win;
	GtkWidget *note;
	Display *disp;
	int major, minor, bank, ram;
	GtkWidget *parent;
	GdkCursor *ptr;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(IS_GTVSCREEN(widget));


	GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);
	gtv = GTVSCREEN(widget);

	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.event_mask = gtk_widget_get_events(widget) |
	    GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
	    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
	    GDK_POINTER_MOTION_HINT_MASK;
	attributes.visual = gtk_widget_get_visual(widget);
	gtv->visual = attributes.visual;

	attributes.colormap = gtk_widget_get_colormap(widget);

	attributes_mask =
	    GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
	widget->window =
	    gdk_window_new(widget->parent->window, &attributes,
			   attributes_mask);
	ptr = gdk_cursor_new(GDK_RTL_LOGO);
	gdk_window_set_cursor(widget->window, ptr);
	gdk_cursor_destroy(ptr);

	widget->style = gtk_style_attach(widget->style, widget->window);

	gdk_window_set_user_data(widget->window, widget);
	gtk_style_set_background(widget->style, widget->window,
				 GTK_STATE_SELECTED);

	disp = (Display *) GDK_DISPLAY();
	win = widget->window;

	gtk_widget_set_usize(GTK_WIDGET(gtv), 0, 0);
	gtv->vwin.width = attributes.width;
	gtv->vwin.height = attributes.height;
	gtv->vwin.x = 0;
	gtv->vwin.y = 0;
	gtv->vbuf.width = gdk_screen_width();
	gtv->vbuf.height = gdk_screen_height();
	gtv->vbuf.base = 0;
	gtv->vwin.clipcount = 0;

	rwidth = gtv->vbuf.width;

	fprintf(stderr, "Screensize:%dx%d\n", gdk_screen_width(),
		gdk_screen_height());
#ifndef NOCARD
#ifndef NODGA
	if (!gtv->remote && XF86DGAQueryVersion(disp, &major, &minor)) {
		fprintf(stderr, "DGA Version %d.%d\n", major, minor);
		switch (major) {
		case 1:
			XF86DGAGetVideoLL(disp, DefaultScreen(disp),
					  (void *) &(gtv->vbuf.base),
					  &rwidth, &bank, &ram);
			break;
		case 2:
			XF86DGAGetVideoLL(disp, DefaultScreen(disp),
					  (void *) &(gtv->vbuf.base),
					  &rwidth, &bank, &ram);
			break;

		default:
			gtv->remote = TRUE;
		}
	} else {
		fprintf(stderr, "Can`t use DGA\n");
		gtv->remote = TRUE;
	}
	fprintf(stderr,"addr: 0x%08x\n",gtv->vbuf.base);


#endif
	if (gtv->realw)
		rwidth = gtv->realw;

	if (gtv->bpp)
		gtv->vbuf.depth = gtv->bpp;
	else {
		Screen *s;
		int ndepths = 0, *depths = NULL;
 		int n,i;
		Display *dpy;
		XPixmapFormatValues *pf;

		
		dpy=XOpenDisplay(NULL);
		s = ScreenOfDisplay (dpy,DefaultScreen (dpy));

		depths = XListDepths (dpy, DefaultScreen (dpy), &ndepths);
		if (!depths) ndepths = 0;
		gtv->vbuf.depth = depths[ndepths-1];

		pf = XListPixmapFormats(dpy,&n);
		i = 0;
		while (i < n) 
		{
			if ( pf[i].depth == gtv->vbuf.depth ){
				gtv->bpp = pf[i].bits_per_pixel;
				break;
			}
			i++;
		}
		
		fprintf(stderr,"Depth: %d\n",gtv->vbuf.depth);
		fprintf(stderr,"BPP: %d\n",gtv->bpp);

		XFree(pf);
		XCloseDisplay(dpy);

	}
	gtv->Format = which_format(gtv->bpp);

	gtv->vbuf.bytesperline =
	    rwidth * (0x38 & (gtv->vbuf.depth + 1)) / 8;
	gtv->vp.depth = gtv->vbuf.depth;
	gtv->vp.palette = which_format(gtv->vbuf.depth);



	if (myioctl(gtv->fbttv, VIDIOCSFBUF, &gtv->vbuf)) {
		perror("set frame buffer");
		exit(1);
	}
	setuid(getuid());
	myioctl(gtv->fbttv, VIDIOCSPICT, &gtv->vp);

#endif



	parent = gtk_widget_get_toplevel (widget);


	gtk_signal_connect_object_after(GTK_OBJECT(parent), 
					"map_event", (GtkSignalFunc)
					pConfig, GTK_OBJECT(widget));
	gtk_signal_connect_object_after(GTK_OBJECT(parent), "unmap_event",
			   (GtkSignalFunc) pConfig, GTK_OBJECT(widget));
	gtk_signal_connect_object_after(GTK_OBJECT(parent), "configure_event",
			   (GtkSignalFunc) pConfig, GTK_OBJECT(widget));


	XSelectInput(disp, DefaultRootWindow(disp), VisibilityChangeMask |
		     SubstructureNotifyMask | ExposureMask |
		     StructureNotifyMask);



	gdk_window_add_filter((GdkWindow *) & gdk_root_parent,
			      (GdkFilterFunc) pConfigFilter, gtv);

	setmaxsize(gtv);

	if (!gtv->image) {
		gtv->image =
		    gdk_image_new(GDK_IMAGE_FASTEST, gtv->visual, MAXX,
				  MAXY+256);
		gtv->gc = gdk_gc_new(widget->window);
	}

	gdk_window_show(widget->window);
	pConfig(gtv, 0, gtv);

	note = CreateNote("Gtvscreen", GTVSCREEN(widget));
	gtk_signal_connect(GTK_OBJECT(widget), "button_press_event",
			   (GtkSignalFunc) button_press, (gpointer) note);

	set_geo(gtv);
	XSetErrorHandler((XErrorHandler) xerror); 
}


static void gtvscreen_init(Gtvscreen * gtv)
{
	GdkColor *my;
	GdkColormap *cmap;
	char buf[16];
	ushort i;
	GdkVisual *v;
	GtkWidget *widget = GTK_WIDGET(gtv);

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));


	GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(gtv), GTK_NO_WINDOW);
	gtv->gc = (GdkGC *) NULL;
	gtv->image = (GdkImage *) NULL;
	gtv->CMSET = FALSE;

	v = gtk_widget_get_visual(widget);
	if (v->type == GDK_VISUAL_PSEUDO_COLOR) {
		fprintf(stderr,
			"Setting colormap for PseudoColor Visual\n");
		my = gtv->colors;
		cmap = gdk_colormap_new(gdk_visual_get_best(), 256);
		for (i = 0; i < 16; i++) {
			gdk_color_black(cmap, &my[i]);
		}
		for (i = 0; i < 225; i++) {
			int r, rr;
			r = i % 25;
			rr = r % 5;
			sprintf(buf, "#%02X%02X%02X",
				(65535 * (r / 5) / 4),
				(65535 * (i / 25) / 8), (65535 * rr / 4));
			gdk_color_parse(buf, &my[i + 16]);
			gdk_color_alloc(cmap, &my[i + 16]);
		}
		gtk_widget_push_colormap(cmap);
		gtv->cmap = cmap;
		gtk_widget_set_colormap(widget, cmap);
	}
	// get picture settings and set reasonable values
	myioctl(gtv->fbttv, VIDIOCGPICT, &gtv->vp);

	gtv->vp.brightness = (128) << 8;
	gtv->vp.colour = (256) << 7;
	gtv->vp.hue = (unsigned int) (128) << 8;
	gtv->vp.contrast = (256) << 7;
	gtv->vp.depth = gtv->bpp;
	gtv->vp.palette = gtv->Format;

	myioctl(gtv->fbttv, VIDIOCSPICT, &gtv->vp);


	// get input settings 
	myioctl(gtv->fbttv, VIDIOCGCAP, &gtv->vc);

	gtv->vc.norm = 0;
	gtv->vc.channel = 0;

	// get window settings 
	myioctl(gtv->fbttv, VIDIOCGWIN, &gtv->vwin);

	gtv->accel_group = gtk_accel_group_new();
	gtk_accel_group_attach(gtv->accel_group, GTK_OBJECT(gtv));


	// get audio settings (and set reasonable values)
	myioctl(gtv->fbttv, VIDIOCGAUDIO, &gtv->va);
	myioctl(gtv->fbttv, VIDIOCSAUDIO, &gtv->va);

	gtv->refresh = 1;
}


static void gtvscreen_destroy(GtkObject * object)
{
	Gtvscreen *gtv;

	g_return_if_fail(object != NULL);
	g_return_if_fail(IS_GTVSCREEN(object));

	gtv = GTVSCREEN(object);

	gtv->CAP_ON = 0;
	set_capture(gtv, 0);
	close(gtv->fbttv);
#ifdef DEBUG
	printf("Widget destroyed\n");
#endif

}


static void set_capture(Gtvscreen * gtv, int W)
{
	static int timer;
	int one = 1;
	int zero = 0;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	if (gtv->remote && gtv->OVL_ON) {
		gtv->OVL_ON = FALSE;
	}
#ifndef NOCARD

	if (GTK_WIDGET_REALIZED(gtv)) {

		switch (W) {
		case 1:
			if (gtv->CAP_ON) {
				if (gtv->OVL_ON) {
					get_clip(gtv);
					set_geo(gtv);
					if (myioctl
					    (gtv->fbttv, VIDIOCCAPTURE,
					     &one)) perror("capture on");
				} else {
					if (myioctl
					    (gtv->fbttv, VIDIOCCAPTURE,
					     &zero)) perror("capture off");
					gtv->start = TRUE;
					gtv->stop = FALSE;
					timer =
					    gtk_idle_add(paint,
							 (gpointer) gtv);
				}
			}
			break;

		case 0:
		default:
			if (timer) {
				gtk_idle_remove(timer);
				gtv->stop = TRUE;
				grab_image(gtv);
				gtv->stop = FALSE;
				timer = 0;
			} else {
				if (myioctl
				    (gtv->fbttv, VIDIOCCAPTURE,
				     &zero)) perror("capture off");
			}

		}
#endif
	}
}

#ifndef NOCARD
static void add_clip(struct video_clip *cr, int *ncr, uint x, uint y,
		    uint w, uint h)
{

	if (*ncr > MAXCLIP) {
		fprintf(stderr, "Too many clipped windows\n");
		return;
	}
	cr[*ncr].x = x;
	cr[*ncr].y = y;
	cr[*ncr].width = w;
	cr[*ncr].height = h;
	*ncr = *ncr + 1;
}

#endif

//#define OSDTEST
#ifdef OSDTEST
unsigned int mask[VIDEO_CLIPMAP_SIZE / 4];
#endif

static void maketab(Gtvscreen * gtv, int ncr)
{
	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

#ifndef NOCARD
#ifdef OSDTEST
	{
		memset(mask, 0xf0, VIDEO_CLIPMAP_SIZE);

		gtv->vwin.flags = VIDEO_CLIP_BITMAP;
		gtv->vwin.clips = (struct video_clip *) mask;
		if (ncr && gtv->OVL_ON && gtv->CAP_ON && gtv->vwin.width
		    && gtv->vwin.height) {
			if (myioctl(gtv->fbttv, VIDIOCSWIN, &(gtv->vwin)))
				perror("set clipping");
		}
#else
	gtv->vwin.clips = gtv->cliprecs;
	gtv->vwin.clipcount = ncr;
	if (ncr && gtv->OVL_ON && gtv->CAP_ON && gtv->vwin.width
	    && gtv->vwin.height) {
		if (myioctl(gtv->fbttv, VIDIOCSWIN, &(gtv->vwin)))
			perror("set clipping");
#endif
	}
#endif
}

#ifndef NOCARD
static int get_clip(Gtvscreen * gtv)
{
	int ncr = 0;
	int x, y, x2, y2;
	Display *disp;
	XWindowAttributes wts;
	Window parent, win, parent2, root, rroot, *children, swin;
	uint nchildren, i;
	GtkWidget *widpar;

	if (!GTK_WIDGET_REALIZED(GTK_WIDGET(gtv))
	    || !gtv->OVL_ON)
		return 0;


	disp = (Display *) GDK_DISPLAY();

	widpar = GTK_WIDGET(gtv);
	while (widpar->parent)
		widpar = widpar->parent;
	if (!(win = GDK_WINDOW_XWINDOW(widpar->window)))
		return 0;

	XQueryTree(disp, win, &rroot, &parent, &children, &nchildren);
	if (nchildren)
		XFree((char *) children);

	root = DefaultRootWindow(disp);
	swin = win;
	while (parent != root) {
		swin = parent;
		XQueryTree(disp, swin, &rroot, &parent, &children,
			   &nchildren);
		if (nchildren)
			XFree((char *) children);
	}
	XQueryTree(disp, root, &rroot, &parent2, &children, &nchildren);
	for (i = 0; i < nchildren; i++)
		if (children[i] == swin)
			break;
	i++;
	for (; i < nchildren; i++) {

		XGetWindowAttributes(disp, children[i], &wts);
		if (!(wts.map_state & IsViewable) || !onscreen(gtv, wts))
			continue;

		x = wts.x - gtv->vwin.x;
		y = wts.y - gtv->vwin.y;
		x2 = wts.width + 2 * wts.border_width;
		y2 = wts.height + 2 * wts.border_width;
		add_clip(gtv->cliprecs, &ncr, x, y, x2, y2);
	}
	XFree((char *) children);


	maketab(gtv, ncr);
	return ncr;
}

#endif


static void set_geo(Gtvscreen * gtv)
{
#ifndef NOCARD
	gint x, y, w, h;
	gint dx, dy;
	Boolean CHECK;
	gint oldx, oldy, oldw, oldh;
	GtkWidget *widget;
	GtkWidget *parent;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	oldx = gtv->vwin.x;
	oldy = gtv->vwin.y;
	oldw = gtv->vwin.width;
	oldh = gtv->vwin.height;

	CHECK = gtv->Change;

	gtv->vwin.flags = 0;
	if (!GTK_WIDGET_REALIZED(GTK_WIDGET(gtv)))
		return;

	widget = GTK_WIDGET(gtv);
	parent = widget;
	gdk_window_get_deskrelative_origin(widget->window, &x, &y);

	w = widget->allocation.width;
	h = widget->allocation.height;

	if (oldh != h)
		CHECK = TRUE;
	if (oldw != w)
		CHECK = TRUE;

	dx = 0;
	dy = 0;
	if (w > gtv->vcap.maxwidth) {
		dx = (w - gtv->vcap.maxwidth) / 2;
		if (oldw != gtv->vcap.maxwidth) {
			CHECK = TRUE;
			gtv->vwin.width = gtv->vcap.maxwidth & ~3;
		}
	} else {
		gtv->vwin.width = w & ~3;
	}

	if (h > gtv->vcap.maxheight) {
		dy = (h - gtv->vcap.maxheight) / 2;
		if (oldh != gtv->vcap.maxheight) {
			CHECK = TRUE;
			gtv->vwin.height = gtv->vcap.maxheight & ~3;
		}
	} else
		gtv->vwin.height = h & ~3;

	if (oldx != x + dx) {
		Move = TRUE;
		CHECK = TRUE;
	}
	if (oldy != y + dy) {
		Move = TRUE;
		CHECK = TRUE;
	}

	if ((gtv->vbuf.depth == 24) && ((x + dx) & 3)) {
		parent->allocation.x =
		    parent->allocation.x - ((x + dx) & 3);
	} else if ((gtv->vbuf.depth == 16) && ((x + dx) & 1)) {
		parent->allocation.x =
		    parent->allocation.x - ((x + dx) & 1);
	} else if ((gtv->vbuf.depth == 15) && ((x + dx) & 1)) {
		parent->allocation.x =
		    parent->allocation.x - ((x + dx) & 1);
	}
	if ((y + dy) & 1) {
		parent->allocation.y =
		    parent->allocation.y - ((y + dy) & 1);
	}

	if (CHECK)
		get_clip(gtv);

	if (CHECK) {
		gtv->vwin.x = x + dx;
		gtv->vwin.y = y + dy;
		if (gtv->CAP_ON && gtv->vwin.width && gtv->vwin.height) {
			if (myioctl(gtv->fbttv, VIDIOCSWIN, &gtv->vwin))
				perror("set clipped window");
		}
	}

	gtv->Change = FALSE;
#endif
}



static GdkFilterReturn pConfigFilter(GdkXEvent * xevent,
				     GdkEvent * event, gpointer data)
{
	Window checkw;
	Gtvscreen *gtv;
	XEvent *xe;
	int type;
	XWindowAttributes wts1;

	xe = (XEvent *) xevent;
	gtv = GTVSCREEN(data);
	type = xe->type;
	checkw = xe->xmap.window;

#ifdef DEBUG
	printf("Filter:%d  Window:%d  %d  %d:\n", xe->type, 
	       checkw,lastrwin,GTK_WIDGET(gtv)->window);
#endif
	
	if (checkw == lastrwin || !gtv->OVL_ON)
		return GDK_FILTER_REMOVE;

	if (gtv->CAP_ON && 
	    type == ConfigureNotify &&
	    GTK_WIDGET_VISIBLE(GTK_WIDGET(gtv)) ) {
		XGetWindowAttributes(GDK_DISPLAY(), checkw, &wts1);
		if ( onscreen(gtv, wts1) ){
			set_capture(gtv, 0);
#ifndef NOCARD
			if (gtv->OVL_ON){
				get_clip(gtv);
				refresh(gtv);
			}
#endif
			set_capture(gtv, 1);
		}
	}
	return GDK_FILTER_REMOVE;
}


static gint pConfig(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	Display *disp;
	static Gtvscreen *gtv = NULL;



	if (!gtv && data && (widget == GTK_WIDGET(data))
	    && (event == (GdkEvent *) NULL)) {
			gtv = (Gtvscreen *)data;
			return TRUE;
	}

	if (gtv->OVL_ON && GDK_WINDOW_XWINDOW(widget->window) == lastrwin)
		return FALSE;

#ifdef DEBUG
	printf("pConfig: %d WidgetWindow:%d\n", event->type,
	       GDK_WINDOW_XWINDOW(widget->window));
#endif
	
	disp = (Display *) GDK_DISPLAY();


	if (gtv->CAP_ON && event->type == GDK_CONFIGURE) {
		set_capture(gtv, 0);
		set_geo(gtv);
		if (gtv->OVL_ON && event->type == GDK_CONFIGURE ){
			refresh(gtv);
		}
		set_capture(gtv, 1);
	}

	if (event->type == GDK_UNMAP || event->type == GDK_DESTROY ||
	    event->type == GDK_NO_EXPOSE) {
		if (gtv->CAP_ON) set_capture(gtv, 0);

	}

	return TRUE;
}

static void setmaxsize(Gtvscreen * gtv)
{
	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));
	myioctl(gtv->fbttv, VIDIOCGCAP, &gtv->vcap);
	if (gtv->vcap.maxwidth> MAXX) 
		gtv->vcap.maxwidth=MAXX; 
	if (gtv->vcap.maxheight> MAXY) 
		gtv->vcap.maxheight=MAXY; 
		
		
}

static void change_freq(Gtvscreen * gtv, freqtype * frt)
{
	struct frontend front;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	if (freqchange(frt, &(gtv->FREQ))) {
		set_capture(gtv, 0);

		if(gtv->dvb){
			myioctl(gtv->fbttv, VIDIOCGFRONTEND, &front);
			if (frt->fre < 11800UL)
				front.freq = (frt->fre - 9750UL) * 1000000UL;
			else
				front.freq = (frt->fre - 10600UL) * 1000000UL;
			front.ttk = (frt->fre < 11800UL) ? 0 : 1;
			front.diseqc = frt->lnb;
			front.srate = frt->sr * 1000;
			front.volt = frt->pol - Vpol;
			front.video_pid = frt->vpid;
			front.audio_pid = frt->apid;
			front.fec = frt->fec;
			front.AFC = 1;
			front.channel_flags = DVB_CHANNEL_FTA;
			
			myioctl(gtv->fbttv, VIDIOCSFRONTEND, &front);
		} else {
			if (myioctl(gtv->fbttv, VIDIOCSFREQ, &frt->fre))
				perror("change frequency");
		}

		if (gtv->CAP_ON)
			set_capture(gtv, 1);
	}
}



#define IOBUFSIZ 6144

#ifdef PTHREADS

int th_play = 0;
int th_rec = 0;
int th_pause = 0;
int th_done = 0;

static gint wait_for_thread(gpointer data)
{
	Gtvscreen *gtv;

	gtv = GTVSCREEN(data);
	if (th_done)
		gtvscreen_set_values(gtv, GTV_STOP, (gpointer) 0);

	return TRUE;
}

static void *play(gpointer data)
{
	int count;
	Gtvscreen *gtv;
	int fd1, fd2;
	int status;
	ringbuffy rbuf;
	
	th_done = 0;
	status = pthread_detach(pthread_self());
	if (status)
		fprintf(stderr, "Thread detach %d \n", status);
	gtv = GTVSCREEN(data);
	fd2 = gtv->fbttv;
	fd1=gtv->fd1;
	
	if (fd1 < 1) {
		fprintf(stderr,"Wrong file descriptor %d\n",fd1);
		pthread_exit(NULL);
	}

	ring_init(&rbuf,IOBUFSIZ+20);
	while (th_play && (count = ring_write_file(&rbuf,fd1, IOBUFSIZ))!=0) {
		while (th_pause);
		count = ring_read_file(&rbuf,fd2,IOBUFSIZ);
	} 
	while ( (count  = ring_rest(&rbuf)))
		ring_read_file(&rbuf,fd2,IOBUFSIZ);

	th_done = 1;

	close(fd1);
	pthread_exit(NULL);
}

static void record(gpointer data)
{
	int count;
	Gtvscreen *gtv;
	int fd1, fd2;
	int status;
	ringbuffy rbuf;

	th_done = 0;
	status = pthread_detach(pthread_self());
	if (status)
		fprintf(stderr, "Thread detach %d \n", status);
	gtv = GTVSCREEN(data);
	fd1 = gtv->fbttv;
	fd2 = gtv->fd2;
	if (fd2 < 1) {
		fprintf(stderr,"Wrong file descriptor %d\n",fd2);
		pthread_exit(NULL);
	}

	ring_init(&rbuf,IOBUFSIZ+20);

	while (th_rec && (count = ring_write_file(&rbuf,fd1,IOBUFSIZ))){
		count = ring_read_file(&rbuf,fd2,IOBUFSIZ);
	}
	while ( (count  = ring_rest(&rbuf)))
		ring_read_file(&rbuf,fd2,IOBUFSIZ);
	th_done = 1;

	close(fd2);
	pthread_exit(NULL);

}


#else
static gint play(gpointer data)
{
	int count;
	char buf[IOBUFSIZ];
	Gtvscreen *gtv;

	gtv = GTVSCREEN(data);

	if (gtv->fd1 < 1){ 
		fprintf(stderr,"Wrong file descriptor %d\n",gtv->fd1);
		gtvscreen_set_values(gtv, GTV_STOP, (gpointer) 0);
		return TRUE;
	}
	if (!gtv->pause) {
		if ((count = read(gtv->fd1, buf, IOBUFSIZ))) {
			write(gtv->fbttv, buf, count);
		} else {
			gtvscreen_set_values(gtv, GTV_STOP, (gpointer) 0);
			gtv->fd1 = 0;
		}
	}
	return TRUE;
}

static gint record(gpointer data)
{
	int count;
	char buf[IOBUFSIZ];
	Gtvscreen *gtv;

	gtv = GTVSCREEN(data);

	if (!gtv->fd2){
		fprintf(stderr,"Wrong file descriptor %d\n",gtv->fd2);
		gtvscreen_set_values(gtv, GTV_STOP, (gpointer) 0);
		return TRUE;
	}	

	if (!gtv->pause) {
		if ((count = read(gtv->fbttv, buf, IOBUFSIZ))) {
	 		write(gtv->fd2, buf, count);
		} else {
			gtvscreen_set_values(gtv, GTV_STOP, (gpointer) 0);
			gtv->fd2 = 0;
		}
	}
	return TRUE;
}

#endif


Boolean gtvscreen_set_values(Gtvscreen * gtv, int val_type, gpointer value)
{
	Boolean pictmod;
	int old, new;
	struct frontend front;
	static int ptimer;
#ifdef PTHREADS
	pthread_t thread1;
#endif
	pictmod = FALSE;

#ifdef DEBUG
	printf("set value: %i   %i\n", val_type, (int) value);
#endif

	switch (val_type) {
#ifndef NOCARD

	case GTV_EXT_SW:
		if (gtv->dvb){
			myioctl(gtv->fbttv, VIDIOCGFRONTEND, &front);
			if (!front.ttk)
				gtv->FREQ.fre = front.freq / 
					1000000UL + 9750UL;
			else
				gtv->FREQ.fre = front.freq / 
					1000000UL + 10600UL;
			gtv->FREQ.lnb = front.diseqc;
			gtv->FREQ.sr = front.srate / 1000;
			gtv->FREQ.pol = front.volt + Vpol;
			gtv->FREQ.vpid = front.video_pid;
			gtv->FREQ.apid = front.audio_pid;
			gtv->FREQ.fec = front.fec;
			gtv->Change = TRUE;
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHFREQ]);
			set_geo(gtv);
		}
		break;

	case GTV_PLAY:
		if (gtv->vcap.type & VID_TYPE_MPEG_DECODER){ 
			gtv->pause = FALSE;
			if (gtv->REC_ON) {
				gtk_signal_emit(GTK_OBJECT(gtv),
						gtvscreen_signals[CHRECORD]);
				return TRUE;
			}

			gtv->fd1 = (int) value;
			gtv->pmode.mode = VID_PLAY_NORMAL;
			gtv->pmode.p1 = gtv->vc.norm;
			myioctl(gtv->fbttv, VIDIOCSPLAYMODE, &gtv->pmode);
			gtv->PLY_ON = TRUE;
#ifdef PTHREADS
			th_play = TRUE;
			th_pause = FALSE;
			if (!ptimer) {
				int status;
				if (!ptimer)
					ptimer =
						gtk_idle_add(wait_for_thread,
							     (gpointer) gtv);
				status =
					pthread_create(&thread1, NULL, 
						       (void *) &play,
						       (void *) gtv);
				if (status)
					fprintf(stderr, "Thread detach %d \n",
						status);
			}
#else
			if (!ptimer)
				ptimer = gtk_idle_add(play, (gpointer) gtv);
#endif
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHPLAY]);
		}
		break;

	case GTV_PAUSE:
		if ((gtv->vcap.type & VID_TYPE_MPEG_DECODER) ||
		    (gtv->vcap.type & VID_TYPE_MPEG_ENCODER) ){ 
			if (gtv->REC_ON) {
				gtk_signal_emit(GTK_OBJECT(gtv),
						gtvscreen_signals[CHRECORD]);
				break;
			}
			gtv->pmode.mode = VID_PLAY_PAUSE;
			gtv->pause = TRUE;
			myioctl(gtv->fbttv, VIDIOCSPLAYMODE, &gtv->pmode);
#ifdef PTHREADS
			th_pause = TRUE;
#endif
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHPAUSE]);
		}
		break;

	case GTV_RECORD:
		gtv->pause = FALSE;
		if (gtv->vcap.type & VID_TYPE_MPEG_ENCODER){ 
			if (gtv->PLY_ON) {
				gtk_signal_emit(GTK_OBJECT(gtv),
						gtvscreen_signals[CHRECORD]);
				return TRUE;
			}

			gtv->fd2 = (int) value;
			gtv->pmode.p1 = gtv->vc.norm;
			gtv->pmode.mode = VID_PLAY_NORMAL;
			gtv->REC_ON = TRUE;
#ifdef PTHREADS
			th_rec = TRUE;
			th_pause = FALSE;
			if (!ptimer) {
				int status;
				if (!ptimer)
					ptimer =
						gtk_idle_add(wait_for_thread,
							     (gpointer) gtv);
				status =
					pthread_create(&thread1, NULL,
						       (void *) &record, 
						       (void *) gtv);
				if (status)
					fprintf(stderr, "Thread detach %d \n",
						status);
			}
#else
			if (!ptimer)
				ptimer = gtk_idle_add(record, (gpointer) gtv);
			record((gpointer) gtv);
#endif
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHRECORD]);
		}
		break;

	case GTV_STOP:
		if ((gtv->vcap.type & VID_TYPE_MPEG_DECODER) ||
		    (gtv->vcap.type & VID_TYPE_MPEG_ENCODER) ){ 
#ifdef PTHREADS
			th_play = FALSE;
			th_rec = FALSE;
			th_pause = FALSE;
			if (!th_done)
				break;
#endif
			gtv->pmode.mode = VID_PLAY_RESET;
			gtv->PLY_ON = FALSE;
			gtv->REC_ON = FALSE;
			myioctl(gtv->fbttv, VIDIOCSPLAYMODE, &gtv->pmode);
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHSTOP]);
			if (ptimer)
				gtk_idle_remove(ptimer);
			ptimer = 0;
			if (gtv->fd1 >=0) close (gtv->fd1);
			if (gtv->fd2 >=0) close (gtv->fd2);
			gtv->fd1 = -1;
			gtv->fd2 = -1;
		}
		break;

	case GTV_FFWD:
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHFFWD]);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHSTOP]);
		break;
	case GTV_FBWD:
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHFBWD]);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHSTOP]);
		break;
	case GTV_FSKP:
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHFSKP]);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHSTOP]);
		break;
	case GTV_BSKP:
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHBSKP]);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHSTOP]);
		break;

	case GTV_CAP:
		old = gtv->CAP_ON;
		gtv->CAP_ON = (int) value;
		if (old != gtv->CAP_ON) {
			switch (gtv->CAP_ON) {
			case 1:
				set_capture(gtv, 0);
				get_clip(gtv);
				gtv->Change = TRUE;
				set_geo(gtv);
				myioctl(gtv->fbttv, VIDIOCSCHAN, &gtv->vc);
				myioctl(gtv->fbttv, VIDIOCSPICT, &gtv->vp);
				set_capture(gtv, 1);
				break;
			case 0:
				set_capture(gtv, 0);
				refresh(gtv);
				break;
			}
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHCAPTURE]);
		}
		break;

	case GTV_OVL:
		if (gtv->vcap.type & VID_TYPE_OVERLAY) {
			old = gtv->OVL_ON;
			gtv->OVL_ON = (int) value;
			if (old != gtv->OVL_ON && gtv->CAP_ON) {
				set_capture(gtv, 0);
				//      get_clip(gtv); // kills in play mode
				gtv->Change = TRUE;
				set_geo(gtv);
				set_capture(gtv, 1);
				gtk_signal_emit(GTK_OBJECT(gtv),
						gtvscreen_signals
						[CHOVERLAY]);
			}
		}
		break;

	case GTV_DEVICE:
		if (!GTK_WIDGET_REALIZED(GTK_WIDGET(gtv)))
			strcpy(gtv->devname, (char *) value);
		break;

	case GTV_BPP:
		if (!GTK_WIDGET_REALIZED(GTK_WIDGET(gtv)))
			if (value)
				gtv->bpp = (int) value;
		break;

	case GTV_NORM:
		old = gtv->vc.norm;
		new = (int) value;
		if (new >= 0 && old != new) {
			gtv->vc.norm = new;
			if (gtv->CAP_ON) {
				gtv->Change = TRUE;
				setmaxsize(gtv);
				set_capture(gtv, 0);
				if (myioctl
				    (gtv->fbttv, VIDIOCSCHAN,
				     &gtv->vc)) perror("change norm");
				set_capture(gtv, 1);
				set_geo(gtv);
			}
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHATTRIB]);
		}
		break;

	case GTV_APID:
		if (gtv->dvb){
			gtv->FREQ.apid = (int) value;
			change_freq(gtv, &(gtv->FREQ));
		}
		break;


	case GTV_FREQ:
		gtv->Change = TRUE;
		change_freq(gtv, (freqtype *) value);
		copyfreq(&(gtv->FREQ), (freqtype *) value);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHFREQ]);
		set_geo(gtv);
		break;

	case GTV_INPUT:
		old = gtv->vc.channel;
		gtv->vc.channel = (int) value;
		if (old != gtv->vc.channel) {
			if (gtv->CAP_ON) {
				gtv->Change = TRUE;
				set_capture(gtv, 0);
				if (myioctl
				    (gtv->fbttv, VIDIOCSCHAN,
				     &gtv->vc)) perror("change input");
				set_geo(gtv);
				set_capture(gtv, 1);
			}
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHINPUT]);
		}
		break;


	case GTV_AUDIO:
		old = gtv->va.mode;
		gtv->va.mode = 1 << ((int) value);
		if (old != gtv->va.mode) {
			if (gtv->CAP_ON) {
				gtv->Change = TRUE;
				set_capture(gtv, 0);
				if (myioctl
				    (gtv->fbttv, VIDIOCSAUDIO,
				     &gtv->va)) perror("change audio");
				set_geo(gtv);
				set_capture(gtv, 1);
			}
			gtk_signal_emit(GTK_OBJECT(gtv),
					gtvscreen_signals[CHAUDIO]);
		}
		break;

	case GTV_BRIGHT:
		gtv->Change = TRUE;
		gtv->vp.brightness = ((int) value + 128) << 8;
		pictmod = TRUE;
		break;

	case GTV_COLOR:
		gtv->Change = TRUE;
		gtv->vp.colour = ((int) value) << 7;
		pictmod = TRUE;
		break;

	case GTV_HUE:
		gtv->Change = TRUE;
		gtv->vp.hue = (unsigned int) ((int) value + 128) << 8;
		pictmod = TRUE;
		break;

	case GTV_CONTRAST:
		gtv->Change = TRUE;
		gtv->vp.contrast = ((int) value) << 7;
		pictmod = TRUE;
		break;

	case GTV_VISNAME:	// maybe Ill fix that some day
		break;

	case GTV_REALW:
		if (!GTK_WIDGET_REALIZED(GTK_WIDGET(gtv)))
			gtv->realw = (int) value;
		return TRUE;
		break;

	case GTV_REMOTE:
		if (!gtv->vcap.type & VID_TYPE_OVERLAY)
			gtv->remote = 1;
		else
			gtv->remote = (Boolean) ((int) value);
		break;

#ifdef HAVE_LIBXXF86VM
	case GTV_FULLS:
		set_capture(gtv, 0);
		gtv->next_vid = toggle_fullscreen(gtv->next_vid, gtv);
		set_capture(gtv, gtv->CAP_ON);
		break;
#endif
	}


	if (pictmod) {
		set_capture(gtv, 0);
		if (myioctl(gtv->fbttv, VIDIOCSPICT, &gtv->vp))
			perror("set picture params");
		set_capture(gtv, 1);
		set_geo(gtv);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[CHATTRIB]);
	}
#endif
	return TRUE;

}


Boolean gtvscreen_get_values(Gtvscreen * gtv, int val_type, gint * value)
{

#ifdef DEBUG
	printf("get value: %i  %i\n", val_type, (int) value);
#endif

	switch (val_type) {
#ifndef NOCARD
		
	case GTV_DVB:
		*value = ((gtv->vcap.type & VID_TYPE_DVB) ? 1:0); 
		break;

	case GTV_COVL:
		*value = ((gtv->vcap.type & VID_TYPE_OVERLAY) ? 1:0); 
		break;

	case GTV_MPGE:
		*value = ((gtv->vcap.type & VID_TYPE_MPEG_ENCODER) ? 1:0); 
		break;

	case GTV_MPGD:
		*value = ((gtv->vcap.type & VID_TYPE_MPEG_DECODER) ? 1:0); 
		break;

	case GTV_TUNE:
		*value = ((gtv->vcap.type & VID_TYPE_TUNER) ? 1:0); 
		break;

	case GTV_MAXX:
		setmaxsize(gtv);
		*value = gtv->vcap.maxwidth; 
		break;

	case GTV_MAXY:
		setmaxsize(gtv);
		*value = gtv->vcap.maxheight;
		break;

	case GTV_FD:
		*value = gtv->fbttv;
		break;

	case GTV_CAP:
		*value = gtv->CAP_ON;
		break;

	case GTV_OVL:
		*value = gtv->OVL_ON;
		break;

	case GTV_BPP:
		*value = gtv->bpp;
		break;

	case GTV_NORM:
		*value = gtv->vc.norm;
		break;

	case GTV_FREQ:
		copyfreq((freqtype *) value, &(gtv->FREQ));
		break;

	case GTV_INPUT:
		*value = gtv->vc.channel;
		break;

	case GTV_BRIGHT:
		*value = ((gtv->vp.brightness) >> 8) - 128;
		break;

	case GTV_COLOR:
		*value = (gtv->vp.colour >> 7);
		break;

	case GTV_HUE:
		*value = (int) (gtv->vp.hue >> 8) - 128;
		break;

	case GTV_CONTRAST:
		*value = gtv->vp.contrast >> 7;
		break;

	case GTV_VISNAME:
		value = (gint *) gtv->visname;
		break;

	case GTV_AUDIO:
		*value = gtv->va.audio;
		break;

	case GTV_REALW:
		*value = gtv->realw;
		break;

	case GTV_REMOTE:
		*value = gtv->remote;
		break;

	case GTV_NINPUT:
		setmaxsize(gtv);
		*value = gtv->vcap.channels;
		break;

	case GTV_NAUDIO:
		*value =
		    (gtv->va.mode & 1) + (gtv->va.mode & 2) / 2 +
		    (gtv->va.mode & 4) / 4 + (gtv->va.mode & 8) / 8;
		break;

#ifdef HAVE_LIBXXF86VM
	case GTV_FULLS:
		*value = gtv->isfull;
		break;
#endif

	}

#endif
	return TRUE;

}


static gint paint(gpointer data)
{
	gint x, y, w, h, d;
	GtkWidget *gwidget;
	Gtvscreen *gtv;

	gtv = GTVSCREEN(data);


	grab_image(gtv);
	gwidget = GTK_WIDGET(gtv);

	if (gtv->gc == NULL)
		return FALSE;

	gdk_window_get_geometry(gwidget->window, &x, &y, &w, &h, &d);
	gdk_window_get_deskrelative_origin(gwidget->window, &x, &y);
	if (w > MAXX)
		w = MAXX;

	if (h > MAXY)
		h = MAXY;

	gdk_draw_image(gwidget->window, gtv->gc, gtv->image, 0, 0,
		       gtv->vwin.x - x, gtv->vwin.y - y, w, h);

	gdk_flush();

	return TRUE;
}





static unsigned int which_format(int bpp)
{

	switch (bpp) {

	case 8:
		return VIDEO_PALETTE_HI240;	/* High 240 cube (BT848) */
		break;

	case 16:
		return VIDEO_PALETTE_RGB565;
		break;

	case 24:
		return VIDEO_PALETTE_RGB24;
		break;

	case 32:
		return VIDEO_PALETTE_RGB32;
		break;

	case 15:
		return VIDEO_PALETTE_RGB555;
		break;

	default:
		perror("Cant find color format");
		exit(1);
		break;
	}
}

static void grab_image(Gtvscreen * gtv)
{

	unsigned char *tm, *ctm, *imem;
	struct video_mmap gb;
	int i, fact, w;
	static int frame;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	imem = gtv->image->mem;
#ifndef NOCARD

	if (gtv->CAP_ON && !gtv->OVL_ON) {
		tm = gtv->TargetMem;

		gb.width = gtv->vwin.width;
		gb.height = gtv->vwin.height;

		gb.format = gtv->Format;

		if (gtv->start) {
			gtv->start = FALSE;
			frame = 0;
			gb.frame = 0;
			myioctl(gtv->fbttv, VIDIOCMCAPTURE, &gb);
			frame ^= 1;
		}

		gb.frame = frame;

		myioctl(gtv->fbttv, VIDIOCMCAPTURE, &gb);
		frame ^= 1;
		gb.frame = frame;
		ctm = tm + gtv->mbuf.offsets[gb.frame];
		myioctl(gtv->fbttv, VIDIOCSYNC, &gb.frame);

		fact = (int) (gtv->bpp / 8.0 + .5);
		w = gb.width * fact;
		for (i = 0; i < gb.height; i++)
			memcpy(imem + i * MAXX * fact, ctm + i * w, w);

#endif
	}
	if (gtv->stop) {
		frame ^= 1;
		gb.frame = frame;
		myioctl(gtv->fbttv, VIDIOCSYNC, &gb.frame);
	}
}




static GtkWidget *newpage(GtkWidget * notebook, char *name)
{
	GtkWidget *label;
	GtkWidget *frame;

	label = gtk_label_new(name);
	gtk_widget_show(label);

	frame = gtk_frame_new(name);
	gtk_widget_show(frame);

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame, label);

	return (frame);
}



static GtkWidget *CreateNotebook(GtkWidget * window, GtkPositionType pos)
{
	GtkWidget *box1;
	GtkWidget *notebook;

	gtk_container_border_width(GTK_CONTAINER(window), 0);

	box1 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), box1);

	notebook = gtk_notebook_new();


	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), pos);

	gtk_box_pack_start(GTK_BOX(box1), notebook, TRUE, TRUE, 0);

	gtk_container_border_width(GTK_CONTAINER(notebook), 10);
	return notebook;
}


enum { OVERLAY, CAPTURE, TV, VID1, VID2, VID3, VID4, VID5, BR, COL, HU, CO,
	PA, NT, SE, AU1, AU2, AU3, AU4, FR, SR, AP, VP, FE, TY, LN,
	PO
};


char *AudioNam[4] = { "MONO", "STEREO", "LANG1", "LANG2" };

static void capture_stuff(GtkWidget * widget, gpointer data)
{
	int in, i;
	int cap;
	int act;
	int l;
	static Gtvscreen *gwidget = NULL;
	freqtype freq;
	char *entry;

	if (!gwidget) {
		g_return_if_fail(data != NULL);
		g_return_if_fail(IS_GTVSCREEN(data));

		gwidget = GTVSCREEN(data);
		return;
	}

	l = (int) data;

	if (l >= FR)
		gtvscreen_get_values(gwidget, GTV_FREQ, (gpointer) & freq);


	if (GTK_WIDGET_VISIBLE(gwidget)) {

		switch (l) {

			//switch on overlay 
		case OVERLAY:
			act =
			    (int) (gtk_toggle_button_get_active
				   ((GtkToggleButton *) widget));
			gtvscreen_get_values(gwidget, GTV_OVL, &cap);
			if (cap != act)
				gtvscreen_set_values(gwidget, GTV_OVL,
						     (gpointer) act);
			break;

			// switch on capturing
		case CAPTURE:
			act =
			    (int) (gtk_toggle_button_get_active
				   ((GtkToggleButton *) widget));
			gtvscreen_get_values(gwidget, GTV_CAP, &cap);
			if (cap != act)
				gtvscreen_set_values(gwidget, GTV_CAP,
						     (gpointer) act);
			break;

			// switch to selected input
		case TV:
		case VID1:
		case VID2:
		case VID3:
		case VID4:
		case VID5:
			act =
			    (int) (gtk_toggle_button_get_active
				   ((GtkToggleButton *) widget));
			if (act) {
				in = l - TV;
				gtvscreen_set_values(gwidget, GTV_INPUT,
						     (gpointer) in);
			}
			break;

			// switch to selected picture adjustment
		case BR:
		case COL:
		case HU:
		case CO:
			act = GTK_ADJUSTMENT(widget)->value;
			in = GTV_BRIGHT - BR + l;
			gtvscreen_set_values(gwidget, in, (gpointer) act);
			break;

		case PA:
		case NT:
		case SE:
			act =
			    (int) (gtk_toggle_button_get_active
				   ((GtkToggleButton *) widget));
			if (act) {
				in = l - PA;
				gtvscreen_set_values(gwidget, GTV_NORM,
						     (gpointer) in);
			}
			break;

		case FR:
			act = GTK_ADJUSTMENT(widget)->value;
			freq.fre = act;
			break;
		case SR:
			act = GTK_ADJUSTMENT(widget)->value;
			freq.sr = act;
			break;
		case FE:
			entry = gtk_entry_get_text(GTK_ENTRY(widget));
			i = 0;
			while (strcmp(entry, combost[i]))
				i++;
			freq.fec = i;
			break;
		case TY:
			entry = gtk_entry_get_text(GTK_ENTRY(widget));
			i = 0;
			while (strcmp(entry, combost[i]))
				i++;
			freq.type = i;
			break;
		case LN:
			entry = gtk_entry_get_text(GTK_ENTRY(widget));
			freq.lnb = atoi(entry);
			break;
		case PO:
			entry = gtk_entry_get_text(GTK_ENTRY(widget));
			i = 0;
			while (strcmp(entry, combost[i]))
				i++;
			freq.pol = i;
			break;
		case AP:
			act = GTK_ADJUSTMENT(widget)->value;
			freq.apid = act;
			break;
		case VP:
			act = GTK_ADJUSTMENT(widget)->value;
			freq.vpid = act;
			break;

		default:
			fprintf(stderr,
				"Error in capture_stuff: Call to %d\n", l);
		}
	}
	if (l >= FR)
		gtvscreen_set_values(gwidget, GTV_FREQ, (gpointer) & freq);

}

static void ovrtoggle(GtkWidget * gtv, gpointer data)
{
	int ovl, ovr;
	GtkWidget *tool = GTK_WIDGET(data);

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	ovr =
	    (int) (gtk_toggle_button_get_active((GtkToggleButton *) tool));
	gtvscreen_get_values(GTVSCREEN(gtv), GTV_OVL, &ovl);

	if (ovr != ovl)
		gtk_toggle_button_set_active((GtkToggleButton *) tool,
					     ovl);
}

static void captoggle(GtkWidget * gtv, gpointer data)
{
	int cap, car;
	GtkWidget *tool = GTK_WIDGET(data);

	gtvscreen_get_values(GTVSCREEN(gtv), GTV_CAP, &cap);
	car =
	    (int) (gtk_toggle_button_get_active((GtkToggleButton *) tool));
	if (car != cap)
		gtk_toggle_button_set_active((GtkToggleButton *) tool,
					     cap);
}

static void ch_input(GtkWidget * gtv, gpointer data)
{
	int in, state;
	GtkWidget **tool = (GtkWidget **) data;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	in = GTVSCREEN(gtv)->vc.channel;
	state = gtk_toggle_button_get_active((GtkToggleButton *) tool[in]);
	if (!state)
		gtk_toggle_button_set_active((GtkToggleButton *) tool[in],
					     TRUE);
}


static void ch_picture(Gtvscreen * gtv, GtkAdjustment * get)
{
	int type, aval;
	int val = get->value;
	type = get->page_increment;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	gtvscreen_get_values(gtv, GTV_BRIGHT + type, &aval);
	if (aval != val) {
		gtk_adjustment_set_value(get, aval);
	}

}

static void ch_norm(GtkWidget * gtv, gpointer data)
{
	int in, state;
	GtkWidget **tool = (GtkWidget **) data;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	in = GTVSCREEN(gtv)->vc.norm;
	state = gtk_toggle_button_get_active((GtkToggleButton *) tool[in]);
	if (!state)
		gtk_toggle_button_set_active((GtkToggleButton *) tool[in],
					     TRUE);
}


static void ch_freq(Gtvscreen * gtv, GtkAdjustment ** tool)
{
	GtkAdjustment *adj;
	freqtype fr;

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	gtvscreen_get_values(gtv, GTV_FREQ, (gpointer) & fr);
	if (GTK_IS_ADJUSTMENT(tool[0])) {

		adj = GTK_ADJUSTMENT(tool[FR - FR]);
		if (fr.fre != adj->value) {
			gtk_adjustment_set_value(adj, fr.fre);
		}
		adj = GTK_ADJUSTMENT(tool[SR - FR]);
		if (fr.sr != adj->value) {
			gtk_adjustment_set_value(adj, fr.sr);
		}
		adj = GTK_ADJUSTMENT(tool[AP - FR]);
		if (fr.apid != adj->value) {
			gtk_adjustment_set_value(adj, fr.apid);
		}
		adj = GTK_ADJUSTMENT(tool[VP - FR]);
		if (fr.vpid != adj->value) {
			gtk_adjustment_set_value(adj, fr.vpid);
		}
	}
}

static void ch_freq2(Gtvscreen * gtv, GtkEntry ** tool)
{
	freqtype fr;
	GtkEntry *entry;
	char c[2];

	g_return_if_fail(gtv != NULL);
	g_return_if_fail(IS_GTVSCREEN(gtv));

	gtvscreen_get_values(gtv, GTV_FREQ, (gpointer) & fr);
	if (GTK_IS_ENTRY(tool[0])) {
		entry = GTK_ENTRY(tool[FE - FE]);
		if (strcmp(gtk_entry_get_text(entry), combost[fr.fec])) {
			gtk_entry_set_text(entry, combost[fr.fec]);
		}

		entry = GTK_ENTRY(tool[TY - FE]);
		if (fr.type < Analog || fr.type > DVBC) {
			fr.type = AUTO;
		}
		if (strcmp(gtk_entry_get_text(entry), combost[fr.type])) {
			gtk_entry_set_text(entry, combost[fr.type]);
		}
		entry = GTK_ENTRY(tool[PO - FE]);
		if (strcmp(gtk_entry_get_text(entry), combost[fr.pol])) {
			gtk_entry_set_text(entry, combost[fr.pol]);
		}

		entry = GTK_ENTRY(tool[LN - FE]);
		c[0] = (char) (48 + fr.lnb);
		c[1] = (char) 0;
		if (strcmp(gtk_entry_get_text(entry), c)) {
			gtk_entry_set_text(entry, c);
		}
	}
}


char *PicNam[4] = { "Brightness", "Color", "Hue", "Contrast" };
int picmin[4] = { -127, 0, -127, 0 };
int picmax[4] = { 127, 511, 127, 511 };

static GtkWidget *CreateNote(char *title, Gtvscreen * gtv)
{
	GtkWidget *window;
	GtkWidget *widget;
	GtkWidget *notebook;
	GtkWidget *box;
	GtkWidget *hbox1;
	GtkWidget *buttonbox;
	GtkWidget *onoff;
	GtkWidget *overlay;
	GtkObject *adj;
	GList *items = NULL;
	GtkTooltips *tooltips;
	struct video_channel vc;
	GSList *group = NULL;
	int ninput, i, val;
	static GtkWidget *intool[10];
	static GtkWidget *intool2[3];
	static GtkAdjustment *intool3[4];
	static GtkEntry *intool4[4];
	freqtype fr;


	tooltips = gtk_tooltips_new();
	capture_stuff(GTK_WIDGET(gtv), gtv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), title);
	//gtk_widget_set_usize (window, 320, 270);
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, TRUE);
	gtk_accel_group_attach(gtv->accel_group, GTK_OBJECT(window));

	notebook = CreateNotebook(window, GTK_POS_TOP);


	/*              */
	/* Capture Page */
	/*              */
	widget = newpage(notebook, "Capture");
	buttonbox = gtk_hbutton_box_new();
	gtk_object_set_data(GTK_OBJECT(window), "buttonbox", buttonbox);
	gtk_container_add(GTK_CONTAINER(widget), buttonbox);

	onoff = gtk_toggle_button_new_with_label("On/Off");
	gtk_object_set_data(GTK_OBJECT(window), "onoff", onoff);
	gtk_container_add(GTK_CONTAINER(buttonbox), onoff);
	gtk_tooltips_set_tip(tooltips, onoff, "Capture On/Off", NULL);
	if (gtv->CAP_ON)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(onoff),
					     TRUE);

	gtk_signal_connect(GTK_OBJECT(onoff), "toggled",
			   GTK_SIGNAL_FUNC(capture_stuff),
			   (gpointer) CAPTURE);
	gtk_signal_connect(GTK_OBJECT(gtv), "chcapture",
			   GTK_SIGNAL_FUNC(captoggle), (gpointer) onoff);
	gtk_signal_emit(GTK_OBJECT(gtv), gtvscreen_signals[CHCAPTURE]);
	gtk_widget_add_accelerator(onoff, "clicked", gtv->accel_group,
				   'o', 0, 0);


	if (!gtv->remote) {
		overlay = gtk_toggle_button_new_with_label("Overlay");
		gtk_object_set_data(GTK_OBJECT(window), "overlay",
				    overlay);
		gtk_container_add(GTK_CONTAINER(buttonbox), overlay);
		gtk_tooltips_set_tip(tooltips, overlay, "Overlay On/Off",
				     NULL);
		if (gtv->CAP_ON == 2)
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON
						     (overlay), TRUE);

		gtk_signal_connect(GTK_OBJECT(overlay), "toggled",
				   GTK_SIGNAL_FUNC(capture_stuff),
				   (gpointer) OVERLAY);
		gtk_signal_connect(GTK_OBJECT(gtv), "choverlay",
				   GTK_SIGNAL_FUNC(ovrtoggle),
				   (gpointer) overlay);
		gtk_signal_emit(GTK_OBJECT(gtv),
				gtvscreen_signals[OVERLAY]);
		gtk_widget_add_accelerator(overlay, "clicked",
					   gtv->accel_group, 'p', 0, 0);
	}

	/*              */
	/* Input Page */
	/*              */
	widget = newpage(notebook, "Input");

	buttonbox = gtk_vbutton_box_new();
	gtk_object_set_data(GTK_OBJECT(window), "buttonbox2", buttonbox);
	gtk_container_add(GTK_CONTAINER(widget), buttonbox);

	gtvscreen_get_values(gtv, GTV_NINPUT, &ninput);

	for (i = 0; i < ninput; i++) {
		vc.channel = i;
		myioctl(gtv->fbttv, VIDIOCGCHAN, &vc);
		intool[i] =
		    gtk_radio_button_new_with_label(group, vc.name);
		widget = intool[i];
		group = gtk_radio_button_group(GTK_RADIO_BUTTON(widget));
		gtk_object_set_data(GTK_OBJECT(window), vc.name, widget);
		gtk_box_pack_start(GTK_BOX(buttonbox), widget, FALSE,
				   FALSE, 0);
		gtk_widget_set_usize(widget, 35, 35);
		gtk_tooltips_set_tip(tooltips, widget, vc.name, NULL);
		gtk_signal_connect(GTK_OBJECT(widget), "clicked",
				   GTK_SIGNAL_FUNC(capture_stuff),
				   (gpointer) TV + i);
		gtk_widget_add_accelerator(widget, "clicked",
					   gtv->accel_group,
					   (char) ((int) '1' + i), 0, 0);
	}

	gtk_signal_connect(GTK_OBJECT(gtv), "chinput",
			   GTK_SIGNAL_FUNC(ch_input), (gpointer) intool);

	gtk_signal_emit(GTK_OBJECT(gtv), gtvscreen_signals[CHINPUT]);


	/*              */
	/* Picture Page */
	/*              */
	widget = newpage(notebook, "Picture");

	box = gtk_table_new(6, 2, FALSE);
	gtk_object_set_data(GTK_OBJECT(window), "box", box);
	gtk_container_add(GTK_CONTAINER(widget), box);

	for (i = 0; i < 4; i++) {
		widget = gtk_label_new(PicNam[i]);
		gtk_object_set_data(GTK_OBJECT(window), PicNam[i], widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, i, i + 1,
				 (GtkAttachOptions) GTK_EXPAND |
				 GTK_FILL, (GtkAttachOptions)
				 GTK_EXPAND | GTK_FILL, 0, 0);

		gtvscreen_get_values(gtv, GTV_BRIGHT + i, &val);
		adj =
		    gtk_adjustment_new(val, picmin[i], picmax[i], 1,
				       (float) i, 1);
		widget = gtk_hscale_new(GTK_ADJUSTMENT(adj));
		gtk_object_set_data(GTK_OBJECT(window), PicNam[i], widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, i, i + 1,
				 (GtkAttachOptions) GTK_EXPAND |
				 GTK_FILL, (GtkAttachOptions)
				 GTK_EXPAND | GTK_FILL, 0, 0);
		gtk_scale_set_draw_value(GTK_SCALE(widget), FALSE);
		gtk_scale_set_digits(GTK_SCALE(widget), 0);
		gtk_tooltips_set_tip(tooltips, widget, PicNam[i], NULL);


		gtk_signal_connect(adj, "value_changed",
				   GTK_SIGNAL_FUNC(capture_stuff),
				   (gpointer) i + BR);

		gtk_signal_connect(GTK_OBJECT(gtv), "chattrib",
				   GTK_SIGNAL_FUNC(ch_picture),
				   (gpointer) adj);
	}

	// Separators
	widget = gtk_hseparator_new();
	gtk_object_set_data(GTK_OBJECT(window), "hseparator1", widget);
	gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 4, 5,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

	widget = gtk_hseparator_new();
	gtk_object_set_data(GTK_OBJECT(window), "hseparator2", widget);
	gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 4, 5,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

	// norm buttons

	widget = gtk_label_new("Norm");
	gtk_object_set_data(GTK_OBJECT(window), "Norm", widget);
	gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 5, 6,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);


	hbox1 = gtk_hbox_new(FALSE, 0);
	gtk_object_set_data(GTK_OBJECT(window), "hbox1", hbox1);
	gtk_table_attach(GTK_TABLE(box), hbox1, 1, 2, 5, 6,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

	group = NULL;
	widget = gtk_radio_button_new_with_label(group, "PAL");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(widget));
	gtk_object_set_data(GTK_OBJECT(window), "PAL", widget);
	gtk_box_pack_start(GTK_BOX(hbox1), widget, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(widget), "clicked",
			   GTK_SIGNAL_FUNC(capture_stuff), (gpointer) PA);
	intool2[VIDEO_MODE_PAL] = widget;
	gtk_widget_add_accelerator(widget, "clicked", gtv->accel_group,
				   '1', GDK_CONTROL_MASK, 0);
	gtk_tooltips_set_tip(tooltips, widget, "TV Norm PAL", NULL);


	widget = gtk_radio_button_new_with_label(group, "NTSC");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(widget));
	gtk_object_set_data(GTK_OBJECT(window), "NTSC", widget);
	gtk_box_pack_start(GTK_BOX(hbox1), widget, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(widget), "clicked",
			   GTK_SIGNAL_FUNC(capture_stuff), (gpointer) NT);
	intool2[VIDEO_MODE_NTSC] = widget;
	gtk_widget_add_accelerator(widget, "clicked", gtv->accel_group,
				   '2', GDK_CONTROL_MASK, 0);
	gtk_tooltips_set_tip(tooltips, widget, "TV Norm NTSC", NULL);

	widget = gtk_radio_button_new_with_label(group, "SECAM");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(widget));
	gtk_object_set_data(GTK_OBJECT(window), "SECAM", widget);
	gtk_box_pack_start(GTK_BOX(hbox1), widget, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(widget), "clicked",
			   GTK_SIGNAL_FUNC(capture_stuff), (gpointer) SE);
	intool2[VIDEO_MODE_SECAM] = widget;
	gtk_widget_add_accelerator(widget, "clicked", gtv->accel_group,
				   '3', GDK_CONTROL_MASK, 0);
	gtk_tooltips_set_tip(tooltips, widget, "TV Norm SECAM", NULL);

	gtk_signal_connect(GTK_OBJECT(gtv), "chattrib",
			   GTK_SIGNAL_FUNC(ch_norm), (gpointer) intool2);


	gtk_signal_emit(GTK_OBJECT(gtv), gtvscreen_signals[CHATTRIB]);

	/*                */
	/* Tune Page      */
	/*                */

	widget = newpage(notebook, "Tune");

	box = gtk_table_new(6, 2, TRUE);
	gtk_object_set_data(GTK_OBJECT(window), "table1", box);
	gtk_container_add(GTK_CONTAINER(widget), box);

	gtvscreen_get_values(gtv, GTV_FREQ, (gpointer) & fr);

	/* Frequency               */
	widget = gtk_label_new("Frequency");
	gtk_object_set_data(GTK_OBJECT(window), "freql", widget);
	gtk_widget_show(widget);
	gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 0, 1,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

	adj = gtk_adjustment_new(fr.fre, 10, 20000, 1, 10, 10);
	widget = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 10, 0);
	gtk_object_set_data(GTK_OBJECT(window), "freq", widget);
	gtk_widget_show(widget);
	gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 1, 2,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
			 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(widget), TRUE);
	gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(widget), TRUE);
	gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(widget), TRUE);
	gtk_tooltips_set_tip(tooltips, widget, "Adjust Frequency", NULL);
	intool3[FR - FR] = GTK_ADJUSTMENT(adj);
	gtk_signal_connect(adj, "value_changed",
			   GTK_SIGNAL_FUNC(capture_stuff), (gpointer) FR);

	if (gtv->dvb){

	/* Signal Rate               */
		widget = gtk_label_new("Signal Rate");
		gtk_object_set_data(GTK_OBJECT(window), "srl", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 2, 3,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0,
				 0);
		
		adj = gtk_adjustment_new(fr.sr, 1000, 30000, 1, 10, 10);
		widget = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 1, 0);
		gtk_object_set_data(GTK_OBJECT(window), "SR", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 3, 4,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(widget), TRUE);
		gtk_tooltips_set_tip(tooltips, widget, "Adjust Signal Rate", 
				     NULL);
		intool3[SR - FR] = GTK_ADJUSTMENT(adj);
		gtk_signal_connect(adj, "value_changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) SR);

		
		
		/* Type               */
		widget = gtk_label_new("Type");
		gtk_object_set_data(GTK_OBJECT(window), "typel", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 0, 1,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);

		widget = gtk_combo_new();
		gtk_object_set_data(GTK_OBJECT(window), "Type", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 1, 2,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		items = NULL;
		items = g_list_append(items, combost[Analog]);
		items = g_list_append(items, combost[DVBS]);
		items = g_list_append(items, combost[DVBT]);
		items = g_list_append(items, combost[DVBC]);
		items = g_list_append(items, combost[AUTO]);
		gtk_combo_set_popdown_strings(GTK_COMBO(widget), items);
		g_list_free(items);
		gtk_tooltips_set_tip(tooltips, GTK_COMBO(widget)->entry,
				     "Choose Receiver Type", NULL);
		gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(widget)->entry), 
				       FALSE);
		intool4[TY - FE] = GTK_ENTRY(GTK_COMBO(widget)->entry);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(widget)->entry), 
				   "changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) TY);
		
		
		/* FEC               */
		widget = gtk_label_new("FEC");
		gtk_object_set_data(GTK_OBJECT(window), "fecl", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 2, 3,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		

		widget = gtk_combo_new();
		gtk_object_set_data(GTK_OBJECT(window), "FEC", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 3, 4,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		items = NULL;
		items = g_list_append(items, combost[F1_2]);
		items = g_list_append(items, combost[F2_3]);
		items = g_list_append(items, combost[F3_4]);
		items = g_list_append(items, combost[F4_5]);
		items = g_list_append(items, combost[F5_6]);
		items = g_list_append(items, combost[F6_7]);
		items = g_list_append(items, combost[F7_8]);
		items = g_list_append(items, combost[F8_9]);
		items = g_list_append(items, combost[AUTO]);
		gtk_combo_set_popdown_strings(GTK_COMBO(widget), items);
		g_list_free(items);
		gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(widget)->entry), 
				       FALSE);
		gtk_tooltips_set_tip(tooltips, widget, "Choose FEC", NULL);
		intool4[FE - FE] = GTK_ENTRY(GTK_COMBO(widget)->entry);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(widget)->entry), 
				   "changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) FE);
		


		widget = gtk_label_new("LNB");
		gtk_object_set_data(GTK_OBJECT(window), "lnb", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 4, 5,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);


		widget = gtk_combo_new();
		gtk_object_set_data(GTK_OBJECT(window), "LNB", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 5, 6,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		items = NULL;
		items = g_list_append(items, "0");
		items = g_list_append(items, "1");
		items = g_list_append(items, "2");
		items = g_list_append(items, "3");
		items = g_list_append(items, "AUTO");
		gtk_combo_set_popdown_strings(GTK_COMBO(widget), items);
		g_list_free(items);
		gtk_tooltips_set_tip(tooltips, widget, "Choose LNB", NULL);
		gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(widget)->entry), 
				       FALSE);
		intool4[LN - FE] = GTK_ENTRY(GTK_COMBO(widget)->entry);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(widget)->entry), 
				   "changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) LN);
		
		
		widget = gtk_label_new("Polarization");
		gtk_object_set_data(GTK_OBJECT(window), "pol", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 4, 5,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		

		widget = gtk_combo_new();
		gtk_object_set_data(GTK_OBJECT(window), "POL", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 5, 6,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		items = NULL;
		items = g_list_append(items, combost[Hpol]);
		items = g_list_append(items, combost[Vpol]);
		gtk_combo_set_popdown_strings(GTK_COMBO(widget), items);
		g_list_free(items);
		gtk_tooltips_set_tip(tooltips, widget, "Set Polarization", NULL);
		gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(widget)->entry), 
				       FALSE);
		intool4[PO - FE] = GTK_ENTRY(GTK_COMBO(widget)->entry);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(widget)->entry), 
				   "changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) PO);
		

		widget = gtk_label_new("Audio PID");
		gtk_object_set_data(GTK_OBJECT(window), "apid", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 6, 7,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);

		adj = gtk_adjustment_new(8192, 0, 8192, 1, 10, 10);
		widget = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 1, 0);
		gtk_object_set_data(GTK_OBJECT(window), "APID", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 0, 1, 7, 8,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(widget), TRUE);
		gtk_tooltips_set_tip(tooltips, widget, "Choose Audio PID", 
				     NULL);
		intool3[AP - FR] = GTK_ADJUSTMENT(adj);
		
		gtk_signal_connect(adj, "value_changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) AP);



		widget = gtk_label_new("Video PID");
		gtk_object_set_data(GTK_OBJECT(window), "vpid", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 6, 7,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);

		adj = gtk_adjustment_new(8192, 0, 8192, 1, 10, 10);
		widget = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 1, 0);
		gtk_object_set_data(GTK_OBJECT(window), "VPID", widget);
		gtk_widget_show(widget);
		gtk_table_attach(GTK_TABLE(box), widget, 1, 2, 7, 8,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
				 (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 
				 0);
		gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(widget), TRUE);
		gtk_tooltips_set_tip(tooltips, widget, "Choose Audio PID",
				     NULL);
		intool3[VP - FR] = GTK_ADJUSTMENT(adj);


		gtk_signal_connect(adj, "value_changed",
				   GTK_SIGNAL_FUNC(capture_stuff), 
				   (gpointer) VP);
		
		
		gtk_signal_connect(GTK_OBJECT(gtv), "chfreq",
				   GTK_SIGNAL_FUNC(ch_freq), 
				   (gpointer) intool3);

		gtk_signal_connect(GTK_OBJECT(gtv), "chfreq",
				   GTK_SIGNAL_FUNC(ch_freq2), 
				   (gpointer) intool4);


	}
	
	return window;
}

static gint button_press(GtkWidget * widget, GdkEventButton * event,
			 GtkWidget * note)
{

	if (event->button == 3) {
		if (GTK_WIDGET_VISIBLE(note))
			gtk_widget_hide(note);
		else
			gtk_widget_show_all(note);
	}


	return TRUE;
}

#ifdef HAVE_LIBXXF86VM
static int toggle_fullscreen(int nmode, Gtvscreen * gtv)
{
	int i, cmode, dummy, width, height, x, y, dx, dy;
	Display *disp;
	XF86VidModeModeLine modeline;
	GtkWidget *parent, *widget;
	GtkAllocation alloc;
	int found = FALSE;

	static XF86VidModeModeInfo **modelines;
	static int full;
	static int vmodes = 0;
	static int to, in, pb, aex;
	GdkCursor *ptr;
	GdkPixmap *pixi;
	char data[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

	disp = (Display *) GDK_DISPLAY();
	widget = GTK_WIDGET(gtv);

	if (gtv->remote)
		return -1;
	if (!vmodes) {
		XF86VidModeGetAllModeLines(disp, DefaultScreen(disp),
					   &vmodes, &modelines);


		for (i = 0; i < vmodes; i++)
			if (FULLX == modelines[i]->hdisplay &&
			    FULLY == modelines[i]->vdisplay) {
				found = TRUE;
				break;
			}
		full = i;

		XF86VidModeGetModeLine(disp, DefaultScreen(disp), &dummy,
				       &modeline);
		for (i = 0; i < vmodes; i++)
			if (modeline.hdisplay == modelines[i]->hdisplay
			    && modeline.vdisplay == modelines[i]->vdisplay)
				break;
		cmode = i;

		if (!found) {
			fprintf(stderr,
				"Couldn't find fullscreen modeline %dx%d\n",
				FULLX, FULLY);
			fprintf(stderr,
				"Setting fullscreen to current mode\n");
			full = cmode;
		}

		nmode = full;
		gtv->isfull = TRUE;
	} else {
		XF86VidModeGetModeLine(disp, DefaultScreen(disp), &i,
				       &modeline);
		for (i = 0; i < vmodes; i++)
			if (modeline.hdisplay == modelines[i]->hdisplay
			    && modeline.vdisplay == modelines[i]->vdisplay)
				break;
		cmode = i;
		if (gtv->isfull)
			gtv->isfull = FALSE;
		else
			gtv->isfull = TRUE;
	}
	if (nmode != cmode)
		XF86VidModeSwitchToMode(disp, DefaultScreen(disp),
					modelines[nmode]);

	if (gtv->isfull) {
		width = modelines[full]->htotal;
		height = modelines[full]->vtotal;

		parent = widget->parent;
		while (parent->parent) {
			parent = parent->parent;
		}

		alloc.width = FULLX;
		alloc.height = FULLY;
		alloc.x = 0;
		alloc.y = 0;
		gtk_widget_size_allocate(parent, &alloc);
		gdk_window_move(parent->window, 0, 0);

		gdk_window_get_deskrelative_origin(widget->window, &x, &y);
		dx = parent->allocation.width - gtv->vwin.width;
		dy = parent->allocation.height - gtv->vwin.height;
		gdk_window_move(parent->window, dx / 2- x - 1,
				dy / 2 - y - 5 - 1);
		alloc.width = FULLX;
		alloc.height = FULLY;
		alloc.x = 0;
		alloc.y = 0;
		gtk_widget_size_allocate(widget, &alloc);


		XF86VidModeLockModeSwitch(disp, DefaultScreen(disp), TRUE);
		XWarpPointer(disp, None,
			     RootWindow(disp, DefaultScreen(disp)), 0, 0,
			     0, 0, 0, 0);

		XGetScreenSaver(disp, &to, &in, &pb, &aex);
		XSetScreenSaver(disp, 0, 0, DefaultBlanking,
				DefaultExposures);

		parent = widget->parent;
		while (parent->parent) {
			gtk_widget_hide(parent);
			parent = parent->parent;
		}
		pixi =
		    gdk_bitmap_create_from_data(widget->window, data, 8,
						8);
		ptr =
		    gdk_cursor_new_from_pixmap(pixi, pixi, gtv->colors,
					       gtv->colors, 0, 0);
		gdk_window_set_cursor(widget->window, ptr);
		gdk_cursor_destroy(ptr);
		gdk_pixmap_unref(pixi);


		return cmode;
	} else {
		parent = widget;
		while (parent->parent) {
			parent = parent->parent;
			gtk_widget_show(parent);
		}
		gdk_window_move(parent->window, 30, 30);
		ptr = gdk_cursor_new(GDK_RTL_LOGO);
		gdk_window_set_cursor(widget->window, ptr);
		gdk_cursor_destroy(ptr);

		gtk_widget_show(parent);
		XF86VidModeLockModeSwitch(disp, DefaultScreen(disp),
					  FALSE);
		XSetScreenSaver(disp, to, in, pb, aex);
		return full;
	}

}

#endif
